defmodule Mix.Tasks.Day8 do
  alias AoC.Util.Math
  use Mix.Task
  defp instruction_to_index("L"), do: 0
  defp instruction_to_index("R"), do: 1

  def walk(_, _, step, <<_::binary-size(2)>> <> "Z") do
    step
  end

  def walk(instructions, nodes, step, current_location) do
    index =
      instructions
      |> String.at(rem(step, String.length(instructions)))
      |> instruction_to_index()

    new_location = elem(Map.get(nodes, current_location), index)
    walk(instructions, nodes, step + 1, new_location)
  end

  def loop_receive(max_len, acc, count) when max_len == count do
    acc
  end

  def loop_receive(max_len, acc, count) do
    receive do
      n -> loop_receive(max_len, AoC.Util.Math.lcm(acc, n), count + 1)
    end
  end

  def to_rooms(
        <<from::binary-size(3)>> <>
          " = (" <> <<a::binary-size(3)>> <> ", " <> <<b::binary-size(3)>> <> ")"
      ) do
    {from, {a, b}}
  end

  def solve(input, _) do
    [instructions, nodes] = input |> String.split("\n\n")

    nodes =
      nodes
      |> String.split("\n")
      |> Stream.reject(&AoC.Util.String.empty?/1)
      |> Stream.map(&String.trim/1)
      |> Stream.map(&to_rooms/1)
      |> Map.new()

    nodes
    |> Map.keys()
    |> Stream.filter(&(&1 |> String.ends_with?("A")))
    |> Task.async_stream(fn entry -> walk(instructions, nodes, 0, entry) end)
    |> Enum.reduce(
      1,
      fn
        {:ok, e}, acc -> Math.lcm(acc, e)
      end
    )
  end

  def run(_) do
    {_, puzzle_input} = File.read("./input/2023_8.txt")
    solve(puzzle_input, :part2) |> IO.puts()
  end
end
