defmodule AoC.Util.Math do
  @moduledoc """
  Math utility methods.
  """

  @doc """
  Returns the greatest common divisor of integers a and b.
  """
  @spec gcd(integer(), integer()) :: integer()
  def gcd(0, 0) do
    0
  end

  @spec gcd(integer(), integer()) :: integer()
  def gcd(a, a) do
    a
  end

  @spec gcd(integer(), integer()) :: integer()
  def gcd(a, b) when a > b do
    gcd(a - b, b)
  end

  @spec gcd(integer(), integer()) :: integer()
  def gcd(a, b) do
    gcd(a, b - a)
  end

  @doc """
  Returns the least common multiple of integers a and b.
  """
  @spec lcm(integer(), integer()) :: integer()
  def lcm(a, b) do
    div(a * b, gcd(a, b))
  end

  @doc """
  Returns the facorial of n.

  iex> AoC.Util.Math.factorial(0)
  1
  iex> AoC.Util.Math.factorial(1)
  1
  iex> AoC.Util.Math.factorial(2)
  2
  iex> AoC.Util.Math.factorial(5)
  120
  iex> AoC.Util.Math.factorial(-1)
  nil
  """
  def factorial(0), do: 1
  def factorial(1), do: 1
  def factorial(n) when n < 0, do: nil
  def factorial(n), do: factorial(n-1, n)

  defp factorial(1, k), do: k
  defp factorial(n, k), do: factorial(n-1, k*n)
end
